import datetime
from enum import Enum

from cryptography.hazmat.primitives import hashes
from cryptography import x509


class CaType(Enum):
    ROOT = 'ROOT'
    SUBORDINATE = 'SUBORDINATE'


class CaStatus(Enum):
    ISSUE = 'ISSUE'  # 表示正常签发
    REVOKE = 'REVOKE'  # 表示已被吊销


class KeyAlgorithm(Enum):
    RSA2048 = 'RSA2048'
    RSA3072 = 'RSA3072'
    RSA4096 = 'RSA4096'
    EC256 = 'EC256'  # ECDSA with key size 256
    EC384 = 'EC384'  # ECDSA with key size 384
    SM2 = 'SM2'

    def get_key_size(self):
        return {
            'RSA2048': 2048,
            'RSA4096': 4096,
            'EC256': 256,
            'EC384': 384,
            'SM2': 256,
        }[self.value]


class SignatureAlgorithm(Enum):
    SHA256 = 'SHA256'
    SHA384 = 'SHA384'
    SHA512 = 'SHA512'
    SM3 = 'SM3'

    def to_alg(self):
        return {
            'SHA256': hashes.SHA256(),
            'SHA384': hashes.SHA384(),
            'SHA512': hashes.SHA512(),
        }[self.value]


class KeyUsage(Enum):
    DIGITAL_SIGNATURE = 'digitalSignature'
    NON_REPUDIATION = 'nonRepudiation'
    KEY_ENCIPHERMENT = 'keyEncipherment'
    DATA_ENCIPHERMENT = 'dataEncipherment'
    KEY_AGREEMENT = 'keyAgreement'
    KEY_CERT_SIGN = 'keyCertSign'
    CUL_SIGN = 'cRLSign'
    ENCIPHER_ONLY = 'encipherOnly'
    DECIPHER_ONLY = 'decipherOnly'


class ValidityType(Enum):
    """证书有效期类型"""
    YEAR = 'YEAR'  # 年（12个月）
    MONTH = 'MONTH'  # 月（统一按31天）
    DAY = 'DAY'  # 日
    HOUR = 'HOUR'  # 小时


class DistinguishedName(object):
    def __init__(self, country: str = None, state: str = None, locality: str = None, organization: str = None,
                 organization_unit: str = None, common_name: str = None):
        """
        证书名称配置
        :param country: 国家编码
        :param state: 省市名称
        :param locality: 地区名称
        :param organization: 组织名称
        :param organization_unit: 组织单元名称
        :param common_name: 证书通用名称
        """
        self.country = country
        self.state = state
        self.locality = locality
        self.organization = organization
        self.organization_unit = organization_unit
        self.common_name = common_name

    def build_subject(self) -> x509.Name:
        name_attrs = [
            x509.NameAttribute(x509.NameOID.COMMON_NAME, self.common_name)
        ]
        if self.country:
            name_attrs.append(x509.NameAttribute(x509.NameOID.COUNTRY_NAME, self.country))
        if self.state:
            name_attrs.append(x509.NameAttribute(x509.NameOID.STATE_OR_PROVINCE_NAME, self.state))
        if self.locality:
            name_attrs.append(x509.NameAttribute(x509.NameOID.LOCALITY_NAME, self.state))
        if self.organization:
            name_attrs.append(x509.NameAttribute(x509.NameOID.ORGANIZATION_NAME, self.organization))
        if self.organization_unit:
            name_attrs.append(x509.NameAttribute(x509.NameOID.ORGANIZATIONAL_UNIT_NAME, self.organization_unit))
        return x509.Name(name_attrs)


class Validity(object):
    def __init__(self, type: ValidityType = None, value: int = None, start_from: int = None):
        """

        :param type: 有效期类型，YEAR(年，12个月)/MONTH(月，31天)/DAY(天)/HOUR(小时)
        :param value:证书有效期值
        :param start_from: 起始时间，时间戳
        """
        self.type = type
        self.value = value
        self.start_from = start_from

    def get_effective_time(self):
        switcher = {
            ValidityType.YEAR: 12 * 31 * 24,
            ValidityType.MONTH: 31 * 24,
            ValidityType.DAY: 24,
            ValidityType.HOUR: 1
        }

        return datetime.timedelta(hours=switcher.get(self.type) * self.value)


class ExtendKeyUsage(object):
    def __init__(self, server_auth: bool = None, client_auth: bool = None, code_signing: bool = None,
                 email_protection: bool = None, time_stamping: bool = None):
        """
        :param server_auth: 服务器身份验证，OID为：1.3.6.1.5.5.7.3.1。服务器证书请启用本增强型密钥用法，默认值：False
        :param client_auth: 客户端身份验证，OID为：1.3.6.1.5.5.7.3.2。客户端证书请启用本增强型密钥用法，默认值：False
        :param code_signing: 代码签名，OID为：1.3.6.1.5.5.7.3.3。签署可下载的可执行代码客户端认证，默认值：False
        :param email_protection: 安全电子邮件，OID为：1.3.6.1.5.5.7.3.4。电子邮件保护，默认值：False
        :param time_stamping: 时间戳，OID为：1.3.6.1.5.5.7.3.8。将一个对象的哈希绑定到一个时间，默认值：False
        """
        self.server_auth = server_auth
        self.client_auth = client_auth
        self.code_signing = code_signing
        self.email_protection = email_protection
        self.time_stamping = time_stamping

    def build(self):
        key_usages = []
        if self.server_auth:
            key_usages.append(x509.OID_SERVER_AUTH)

        if self.client_auth:
            key_usages.append(x509.OID_CLIENT_AUTH)

        if self.code_signing:
            key_usages.append(x509.OID_CODE_SIGNING)

        if self.email_protection:
            key_usages.append(x509.OID_EMAIL_PROTECTION)

        if self.time_stamping:
            key_usages.append(x509.OID_TIME_STAMPING)


class PrivateCertificateAuthority(object):
    def __init__(self, id=None, type: CaType = None, status: CaStatus = None, key_algorithm: KeyAlgorithm = None,
                 signature_algorithm: SignatureAlgorithm = None, distinguished_name: DistinguishedName = None,
                 not_before=None, not_after=None, created_at=None, updated_at=None, uri=None, password=None):
        self.id = id
        self.type = type
        self.status = status
        self.key_algorithm = key_algorithm
        self.signature_algorithm = signature_algorithm
        self.distinguished_name = distinguished_name
        self.not_before = not_before
        self.not_after = not_after
        self.created_at = created_at
        self.updated_at = updated_at
        self.uri = uri
        self.password = password

    @staticmethod
    def from_db(ca_entity):
        dn = DistinguishedName(country=ca_entity.country,
                               state=ca_entity.state,
                               locality=ca_entity.locality,
                               organization=ca_entity.organization,
                               organization_unit=ca_entity.organization_unit,
                               common_name=ca_entity.common_name)
        return PrivateCertificateAuthority(id=ca_entity.id,
                                           type=CaType(ca_entity.type),
                                           status=CaStatus(ca_entity.status),
                                           key_algorithm=KeyAlgorithm(ca_entity.key_algorithm),
                                           signature_algorithm=SignatureAlgorithm(ca_entity.signature_algorithm),
                                           distinguished_name=dn,
                                           not_before=ca_entity.not_before,
                                           not_after=ca_entity.not_after,
                                           created_at=ca_entity.created_at,
                                           updated_at=ca_entity.updated_at,
                                           uri=ca_entity.uri,
                                           password=ca_entity.password)


class CertificateStatus(Enum):
    ISSUE = 'ISSUE'  # 表示正常签发
    REVOKE = 'REVOKE'  # 表示已被吊销


class PrivateCertificate(object):
    def __init__(self, id=None, status: CertificateStatus = None, issuer_id=None, key_algorithm: KeyAlgorithm = None,
                 signature_algorithm: SignatureAlgorithm = None, distinguished_name: DistinguishedName = None,
                 not_before=None, not_after=None, created_at=None, updated_at=None, uri=None, password=None):
        self.id = id
        self.status = status
        self.issuer_id = issuer_id
        self.key_algorithm = key_algorithm
        self.signature_algorithm = signature_algorithm
        self.distinguished_name = distinguished_name
        self.not_before = not_before
        self.not_after = not_after
        self.created_at = created_at
        self.updated_at = updated_at
        self.uri = uri
        self.password = password

    @staticmethod
    def from_db(cert_entity):
        dn = DistinguishedName(country=cert_entity.country,
                               state=cert_entity.state,
                               locality=cert_entity.locality,
                               organization=cert_entity.organization,
                               organization_unit=cert_entity.organization_unit,
                               common_name=cert_entity.common_name)
        return PrivateCertificate(id=cert_entity.id,
                                  status=CertificateStatus(cert_entity.status),
                                  issuer_id=cert_entity.issuer_id,
                                  key_algorithm=KeyAlgorithm(cert_entity.key_algorithm),
                                  signature_algorithm=SignatureAlgorithm(cert_entity.signature_algorithm),
                                  distinguished_name=dn,
                                  not_before=cert_entity.not_before,
                                  not_after=cert_entity.not_after,
                                  created_at=cert_entity.created_at,
                                  updated_at=cert_entity.updated_at,
                                  uri=cert_entity.uri,
                                  password=cert_entity.password)


class CertificateResourceType(Enum):
    CA = 'CA'
    CERTIFICATE = 'CERTIFICATE'


class CertificateResourceTag(Enum):
    CERTIFICATE = 'CERTIFICATE'
    PRIVATE_KEY = 'PRIVATE_KEY'


class CertificateResource(object):
    def __init__(self, certificate_type: CertificateResourceType,
                 certificate_data: bytes,
                 private_key_data: bytes = None):
        self.certificate_type = certificate_type
        self.certificate_data = certificate_data
        self.private_key_data = private_key_data


class CertificateContent(object):
    def __init__(self, certificate=None, private_key=None, certificate_chain=None):
        self.certificate = certificate
        self.private_key = private_key
        self.certificate_chain = certificate_chain


class DownloadType(Enum):
    APACHE = 'APACHE',
    NGINX = 'NGINX',
    IIS = 'IIS',
    TOMCAT = 'TOMCAT'
    OTHER = 'OTHER'
